import struct
import sys
import json


class NativeMessaging:
    """
    Class for sending native messages using native messaging protocol
    https://developer.mozilla.org/en-US/docs/Mozilla/Add-ons/WebExtensions/Native_messaging
    """
    @staticmethod
    def get_message(fd):
        raw_length = fd.read(4)
        if len(raw_length) == 0:
            sys.exit(0)
        message_length = struct.unpack("@I", raw_length)[0]
        message = fd.read(message_length).decode("utf-8")
        return json.loads(message)

    # Encode a message for transmission, given its content.

    @staticmethod
    def encode_message(message_content):
        encoded_content = json.dumps(message_content).encode("utf-8")
        encoded_length = struct.pack("@I", len(encoded_content))
        return {"length": encoded_length, "content": encoded_content}

    # Send an encoded message to stdout.

    @staticmethod
    def send_message(fd, encoded_message):
        fd.write(encoded_message["length"])
        fd.write(encoded_message["content"])
        fd.flush()



import sys
import pathlib
import time
from evpn import ExpressVpnApi


LIB_PATH = str(pathlib.Path(__file__).parent.absolute() / "..")
sys.path.insert(0, LIB_PATH)
# pylint: disable=wrong-import-position
api = ExpressVpnApi(debug=True)


def test_start_vpn():
    api.start_express_vpn()


def test_connect():
    api.connect("160")
    api.wait_for_connection(timeout=60)
    assert api.is_connected is True


def test_is_running():
    print(api.express_vpn_running)


def test_locations():
    locations = api.locations
    print(locations)


def test_disconnect():
    api.disconnect()
    time.sleep(1)
    assert not api.is_connected


def test_get_status():
    res = api.get_status()
    print(res)


def pytest_sessionfinish(_session, _exitstatus):
    """ whole test run finishes. """
    api.close()



from subprocess import Popen, PIPE
from abc import abstractmethod
import time
import json
from tempfile import gettempdir
import psutil
from .native_messaging import NativeMessaging
from .messages import MessagesV2


class BaseApi:
    """
        Class For Daemon API
        Platform specific methods are abstract
    """

    EXTENSION_ID = "fgddmllnllkalaagkghckoinaemmogpe"
    MESSAGE_API = NativeMessaging()

    def __init__(self, debug=False) -> None:
        self._messages = MessagesV2
        self._locations = None
        self._debug = debug
        self._start_service()
        self._debug_print("Connecting To Daemon")
        self._wait_for_daemon(timeout=5)
        self._debug_print("Connected To Daemon")

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    def _wait_for_daemon(self, timeout):
        connected = False
        start_t = time.time()
        while not connected and time.time() - start_t < timeout:
            message = self._get_message()
            connected = message.get("connected")
            time.sleep(0.3)
        if not connected:
            raise TimeoutError("Can't connect to ExpressVPN daemon")

    def _debug_print(self, data):
        if self._debug:
            print(data[:100] + "...")

    def _build_request(self, method, params=None):
        return {
            "jsonrpc": "2.0",
            "method": method,
            "params": params or {},
            "id": 200
        }

    def _get_message(self):
        message = self.MESSAGE_API.get_message(self._proc.stdout)
        self._debug_print(f"Got message: {json.dumps(message)}")
        return message

    def _send_message(self, message):
        self._debug_print("Sending: " + json.dumps(message))
        self._proc.stdout.flush()  # Flush output first
        self.MESSAGE_API.send_message(
            self._proc.stdin, self.MESSAGE_API.encode_message(message))

    def _get_response(self):
        while True:
            message = self.MESSAGE_API.get_message(self._proc.stdout)
            self._debug_print(f"Got message: {json.dumps(message)}")
            if message.get("type") in ("method", "result") or not message.get("name"):
                return message

    def _call(self, message: str, params=None):
        """ Send native message and return response """
        req = self._build_request(message, params)
        self._send_message(req)
        return self._get_response()

    def _start_service(self):
        path = self._service_path.absolute()
        # pylint: disable-next=consider-using-with
        self._proc = Popen([
            path,
            f"chrome-extension://{self.EXTENSION_ID}/"],
            stdout=PIPE, stdin=PIPE, stderr=PIPE,
            cwd=gettempdir()
        )

    def _get_locations(self):
        return self._call(self._messages.get_locations)

    @property
    @abstractmethod
    def _program_proc_name(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def _program_path(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def _service_path(self):
        raise NotImplementedError

    def start_express_vpn(self):
        path = self._program_path
        # Since in some platforms this process started in attached mode, we keep it running...
        # pylint: disable=consider-using-with
        Popen([path], start_new_session=True)

    @property
    def express_vpn_running(self):
        proc_names = [p.name() for p in psutil.process_iter()]
        return any(p in proc_names for p in self._program_proc_name)

    @property
    def is_connected(self):
        status = self.get_status()
        return bool(status.get("info").get("connected"))

    def get_location_id(self, name):
        found = next(
            (l for l in self._locations if l["name"].lower() == name.lower()), None)
        if not found:
            similar = next(
                (l for l in self._locations if name.lower() in l["name"].lower()), None)
            error_msg = f"Country {name} not found. "
            error_msg += f'Did you mean {similar.get("name")}?' if similar else ""
            raise ValueError(error_msg)
        return found["id"]

    def wait_for_connection(self, timeout=30):
        """
        Wait for connection after calling connect
        Raising TimeoutError if timeout reached
        Default timeout is 30 seconds
        """
        start = time.time()
        while time.time() - start <= timeout:
            if self.is_connected:
                return
            time.sleep(0.3)
        raise TimeoutError

    def wait_for_disconnect(self, timeout=30):
        """
        Wait for connection after calling connect
        Raising TimeoutError if timeout reached
        Default timeout is 30 seconds
        """
        start = time.time()
        while time.time() - start <= timeout:
            if not self.is_connected:
                return
            time.sleep(0.3)
        raise TimeoutError

    def disconnect(self):
        return self._call(self._messages.disconnect)

    def get_status(self):
        return self._call(self._messages.get_status)

    def connect(self, country_id):
        params = {"id": country_id,
                  "change_connected_location": self.is_connected}
        return self._call(self._messages.connect, params)

    def close(self):
        time.sleep(1.5)
        self._proc.kill()

    @property
    @abstractmethod
    def locations(self):
        """
        Get list of locations
        fields: name, ID, country_code
        """
        raise NotImplementedError



from pathlib import Path
from functools import lru_cache
from subprocess import call
from .base_api import BaseApi


class LinuxApi(BaseApi):
    """Class for controlling ExpressVPN daemon on Linux"""

    @property
    def _program_proc_name(self):
        return "ExpressVPN"

    @property
    def _program_path(self):
        return "/usr/bin/expressvpn"

    @property
    def _service_path(self):
        return Path("/usr/bin/expressvpn-browser-helper")

    @property
    @lru_cache
    def locations(self):
        locations = self._get_locations()
        return [
            {
                "id": i["id"],
                "name": i["country"],
                "country_code": i["country_code"]
            } for i in locations["locations"]
        ]

    def start_express_vpn(self):
        stat = call(["systemctl", "start", "--quiet", "expressvpn.service"])
        return stat == 0

    @property
    def express_vpn_running(self):
        stat = call(["systemctl", "is-active",
                    "--quiet", "expressvpn.service"])
        return stat == 0



from pathlib import Path
from functools import lru_cache
from .base_api import BaseApi


class MacApi(BaseApi):
    """Class for controlling ExpressVPN daemon on MacOS"""

    @property
    def _program_proc_name(self):
        return "ExpressVPN"

    @property
    def _program_path(self):
        return "/Applications/ExpressVPN.app/Contents/MacOS/ExpressVPN"

    @property
    def _service_path(self):
        return Path("/Applications/ExpressVPN.app/Contents/MacOS/expressvpn-browser-helper")

    @property
    @lru_cache()
    def locations(self):
        locations = self._get_locations()
        locations = locations["locations"]
        return [
            {
                "id": i["id"], "name": i["country"],
                "country_code": i["country_code"]
            } for i in locations
        ]



# pylint: disable=too-few-public-methods
class MessagesV1:
    """
    Native messaging messages
    For current version on Windows
    """
    get_locations = "GetLocations"
    get_status = "GetStatus"
    connect = "Connect"
    disconnect = "Disconnect"


class MessagesV2:
    """
    Native messaging messages
    For Linux, MacOS and old version on Windows
    """
    get_locations = "XVPN.GetLocations"
    get_status = "XVPN.GetStatus"
    connect = "XVPN.Connect"
    disconnect = "XVPN.Disconnect"



from pathlib import Path
from functools import lru_cache
from .base_api import BaseApi
from .messages import MessagesV1, MessagesV2


class WindowsApi(BaseApi):
    """Class for controlling ExpressVPN daemon on Windows"""

    def __init__(self, debug=False) -> None:
        super().__init__(debug)
        self._messages = MessagesV2() if self._is_old_version else MessagesV1()

    @property
    def _program_proc_name(self):
        return [
            "ExpressVPN.exe",
            "expressvpnd.exe"
        ]

    @property
    def _program_path(self):
        return "C:\\Program Files (x86)\\ExpressVPN\\expressvpn-ui\\ExpressVPN.exe"

    @property
    @lru_cache()
    def _service_path(self):
        paths = [
            Path(
                "C:\\Program Files (x86)\\ExpressVPN\\services\\ExpressVPN.BrowserHelper.exe"),
            Path(
                "C:\\Program Files (x86)\\ExpressVPN\\expressvpnd\\expressvpn-browser-helper.exe")
        ]
        for path in paths:
            if path.exists():
                return path
        raise Exception("Can't find browser service path of expressVPN")

    @property
    @lru_cache()
    def locations(self):
        locs = self._get_locations()
        locs = locs["locations"] if self._is_old_version else locs["data"]["locations"]
        self._locations = [
            {
                "id": i["id"],
                "name": i["name"],
                "country_code": i["country_code"]
            }
            for i in locs
        ]
        return self._locations

    @property
    def _is_old_version(self):
        # lowercase suffix = old version
        return self._service_path.name.endswith('expressvpn-browser-helper.exe')

    @property
    def is_connected(self):
        if self._is_old_version:
            return super().is_connected
        status = self.get_status()
        data = status.get("data")
        if isinstance(data, dict):
            info = data.get("info", {})
            return isinstance(info, dict) and info.get("state") == "connected"
        return False

    def connect(self, country_id):
        if self._is_old_version:
            super().connect(country_id)
        return self._call(self._messages.connect, {"id": country_id})



import platform

_name = platform.system()

if _name == "Windows":
    from .core.windows_api import WindowsApi as ExpressVpnApi
elif _name == "Linux":
    from .core.linux_api import LinuxApi as ExpressVpnApi
elif _name == "Darwin":
    from .core.mac_api import MacApi as ExpressVpnApi
else:
    raise NotImplementedError(f"Can't run on {_name} platform.")


__all__ = ['ExpressVpnApi']
